import asyncio

from .action import AMIAction
from .message import AMIMessage
from .parser import AMIMessageParser
from ..log import agi_ami_logger


class AMIProtocol(asyncio.Protocol):
    encoding = 'ascii'

    def __init__(self, manager):
        self.manager = manager
        self.responses = {}
        self._message_parser = AMIMessageParser(self.encoding)

        self.transport = None

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        self.manager.connection_lost(exc)

    def data_received(self, data):
        messages = self._message_parser.feed_data(data)
        if not messages:
            return

        for message in messages:
            self.handle_message(message)

    def handle_message(self, message: AMIMessage):
        response = self.responses.get(message.action_id)

        if response is not None:
            if response.add_message(message):
                # completed; dequeue
                self.responses.pop(response.id)

        elif 'Event' in message:
            self.manager.dispatch(message)

    def send(self, action: AMIAction):
        self.responses[action.id] = action

        try:
            self.transport.write(action.as_message(self.encoding))
        except Exception:  # pragma: no cover
            agi_ami_logger.exception('Fail to send %r' % action)

        return action.future



from typing import List
from urllib.parse import unquote

from aioagi.ami.message import AMIMessage


class AMIMessageParser:
    sep = b'\r\n\r\n'
    eol = '\r\n'
    has_body = ('Response: Follows', 'Response: Fail')
    quoted_keys = ['result']
    success_responses = ['Success', 'Follows', 'Goodbye']

    def __init__(self, encoding):
        self.encoding = encoding

        self._tail = b''
        self._sep_len = len(self.sep)

    def feed_data(self, data) -> List[AMIMessage]:
        messages = []
        if self._tail:
            data, self._tail = self._tail + data, b''

        data_len = len(data)
        start_pos = 0

        while start_pos < data_len:
            pos = data.find(self.sep, start_pos)
            if pos >= start_pos:
                # line found
                try:
                    messages.append(self.parse_message(data[start_pos:pos]))
                finally:
                    start_pos = pos + self._sep_len

            else:
                self._tail = data[start_pos:]
                break

        return messages

    def parse_message(self, data: bytearray):
        data = data.decode(self.encoding)
        lines = data.split(self.eol)

        headers = {}
        body = ''
        if lines[0].startswith(self.has_body):
            body = lines.pop()
            while not body and lines:
                body = lines.pop()

        for line in lines:
            if ': ' not in line:
                continue

            key, value = line.split(': ', 1)
            value = value.strip()
            if key.lower() in self.quoted_keys:
                value = unquote(value)

            if key in headers:
                if not isinstance(headers[key], list):
                    headers[key] = [headers[key]]

                headers[key].append(value)
            else:
                headers[key] = value

        return AMIMessage(headers, body=body)



import asyncio

from . import utils


class AMIAction(dict):
    """Dict like object to handle actions.
    Generate action IDs for you:

    ..
        >>> utils.IdGenerator.reset('myuuid')

    .. code-block:: python

        >>> action = AMIAction({'Action': 'Status'})
        >>> print(action) # doctest: +NORMALIZE_WHITESPACE
        Action: Status
        ActionID: action/myuuid/1/1

        >>> action = AMIAction({'Action': 'SIPnotify',
        ...                  'Variable': ['1', '2']})
        >>> print(action) # doctest: +NORMALIZE_WHITESPACE
        Action: SIPnotify
        ActionID: action/myuuid/1/2
        Variable: 1
        Variable: 2
    """
    EOL = '\r\n'
    action_id_generator = utils.IdGenerator('action')

    def __init__(self, *args, as_list=False, **kwargs):
        self.as_list = as_list
        super(AMIAction, self).__init__(*args, **kwargs)

        if 'ActionID' not in self:
            self['ActionID'] = self.action_id_generator()

        self.responses = []
        self.future = asyncio.Future()

    def __str__(self):
        action = []
        for k, v in sorted(self.items()):
            if isinstance(v, (list, tuple)):
                action.extend(['%s: %s' % (k, i) for i in v])
            else:
                action.append('%s: %s' % (k, v))
        action.append(self.EOL)
        return self.EOL.join(action)

    @property
    def id(self):
        return self['ActionID']

    def as_message(self, encoding='ascii'):
        return str(self).encode(encoding)

    @property
    def multi(self):
        resp = self.responses[0]
        msg = resp.get('Message', '').lower()
        if resp.get('SubEvent') == 'Start':
            return True
        elif resp.get('EventList') == 'start':
            return True
        elif 'will follow' in msg:
            return True
        elif msg == 'added interface to queue':
            return False
        elif msg.startswith('added') and msg.endswith('to queue'):
            return True
        elif msg == 'originate successfully queued':
            return False
        elif msg.endswith('successfully queued') and getattr(self, 'async') != 'false':
            return True
        elif self.as_list:
            return True
        return False

    @property
    def completed(self):
        resp = self.responses[-1]
        if resp.get('Event', '').endswith('Complete'):
            return True
        elif resp.get('SubEvent') in ('End', 'Exec'):
            return True
        elif resp.get('Response') in ('Success', 'Error', 'Fail', 'Failure'):
            return True
        elif not self.multi:
            return True
        return False

    def add_message(self, message):
        self.responses.append(message)
        multi = self.multi
        if self.completed and not self.future.done():
            if multi and len(self.responses) > 1:
                self.future.set_result(self.responses)
            elif not multi:
                self.future.set_result(self.responses[0])
            else:
                return False
            return True
